;; Generated package description from nasm-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "nasm-mode" "1.1.1.0.20260509.95321" "NASM x86 assembly major mode" '((emacs "24.3")) :commit "fab76d8e092419c341b6240fcc7123975db177e1" :authors '(("Christopher Wellons" . "wellons@nullprogram.com")) :maintainer '("Christopher Wellons" . "wellons@nullprogram.com") :url "https://github.com/skeeto/nasm-mode")
